import numpy as np

def coherence(A):
    """
    Return the mutual coherence of a matrix A.
    """
    if A.shape[1] > 0:
        gram = A.T @ A
        n = gram.shape[0]
        for i in range(n):
            gram[i,i] = 0.
        return np.max(np.abs(gram))
    else:
        return 0

def update_data(Ginv,B,Aw,b,yw,tw,i,A,y,S,Sbar,l1,l2,AtA,gram,Aty_l1):
    """
    Update the reduced problem data when a new indice is relaxed.
    """

    n     = Ginv.shape[0]                       # Curr. number of relaxed atoms
    Si    = np.sum(S[:i])                       # Position of i in S
    Sbari = np.sum(Sbar[:i])                    # Position of i in Sbar
    perm  = np.insert(np.arange(n), Si, -1)     # Perm. last <-> ith row/column

    # ------------------------------------------------------------------------ #
    #                   Values computed to update problem data                 #
    # Value           Expression                             Dimensions        #
    # ------------------------------------------------------------------------ #
    g               = AtA[S,i]                               # |S|      x 0
    q               = gram[i,i]                              # 1        x 0
    Ginv_g          = Ginv @ g                               # |S|      x 0
    d               = 1 / (q - (g @ Ginv_g))                 # 1        x 0
    Ginv_g_d        = d * Ginv_g                             # |S|      x 0
    d_Ginv_g_Ginv_g = np.outer(Ginv_g_d,Ginv_g)              # |S|      x |S|
    Ginv_11         = Ginv + d_Ginv_g_Ginv_g                 # |S|      x |S|
    Ginv_12         = - Ginv_g_d                             # |S|      x |S|
    Bred            = np.delete(B,Sbari,axis=1)              # |S|-1    x |S|-1
    d_B_g_AtA       = d * (B.T @ g + AtA[Sbar,i])            # |Sbar|   x 0 
    B2              = np.delete(d_B_g_AtA,Sbari)             # |Sbar|-1 x 0
    B1              = Bred + np.outer(Ginv_g,B2.T)           # |Sbar|-1 x |S|-1
    b1              = b                                      # |S|      x 0
    b1             += Aty_l1[S] @ d_Ginv_g_Ginv_g            # |S|      x 0
    b1             -= Aty_l1[i] * Ginv_g_d                   # |S|      x 0
    b2              = - Aty_l1[S] @ Ginv_g_d + Aty_l1[i] * d # 1        x 0
    Aw1             = np.outer(A[:,S] @ Ginv_g - A[:,i], B2) # |S|      x |S|
    yw1             = A[:,S] @ b1                            # m        x 0
    yw2             = b2 * A[:,i]                            # 1        x 0
    tw1             = B1.T @ (l1 + l2*b1)                    # |S|      x 0
    tw2             = - B2 * (l1 + l2*b2)                    # 1        x 0

    # ----------------------------- FLOPS count ------------------------------ #
    # Above computatations
    #   Ginv_g          : |S|^2/2               (symmetric matrix)
    #   d               : 2 + |S|
    #   Ginv_g_d        : |S|
    #   d_Ginv_g_Ginv_g : |S|^2/2               (symmetric outer product)
    #   Ginv_11         : |S|^2/2               (sum of symmetric matrices)
    #   Ginv_12         : |S|
    #   d_B_g_AtA       : |S||Sbar| + |Sbar|
    #   B1              : |S||Sbar|
    #   b1              : |S| + 4
    #   b2              : |S| + 3
    #   Aw1             : |S|^2 + m + |S|
    #   yw1             : m|S|
    #   yw2             : 1
    #   tw1             : 3|S|
    #   tw2             : 4
    #     TOTAL  : (5/2)|S|^2 + 2|S||Sbar| + m|S| + 9|S| + |Sbar| + 14
    # ------------------------------------------------------------------------ #

    # Update of Ginv
    Ginv_new = np.vstack((
        np.hstack((np.reshape(Ginv_11,(n,n)), np.reshape(Ginv_12,(n,1)))),
        np.hstack((np.reshape(Ginv_12,(1,n)), np.reshape(d,(1,1))))  
    ))
    Ginv_new = Ginv_new[:,perm][perm,:]

    # Update of B
    B_new = np.vstack((B1, -B2.T))
    B_new = B_new[perm,:]

    # Update of b
    b_new = np.hstack((np.reshape(b1,(n,)), np.reshape(b2,(1,))))
    b_new = b_new[perm]

    # Update of Aw
    Aw_new = np.delete(Aw,Sbari,axis=1)
    Aw_new += Aw1

    # Update of yw and tw
    yw  = y - yw1 - yw2
    tw  = l1 + tw1 + tw2

    return Ginv_new, B_new, Aw_new, b_new, yw, tw