import time
import numpy as np
from srenet.algebra import coherence, update_data
from srenet.enet import primal_value, duality_gap
from srenet.utils import ( 
    StopCrit, 
    ProjgradParams, 
    ProjgradMonitor, 
    validate_inputs,
    print_header,
    print_trace,
    print_footer,
)

def projgrad(x0, A, y, l1, l2, params=ProjgradParams()):
    """
    Solve the non-negative Elastic-Net problem
        min_{x>=0}  (1/2) ||y-Ax||_2^2 + l1 ||x||_1 + (l2/2) ||x||_2^2
    using a projected-gradient algorithm. Can be overloaded with screening and
    relaxing tests.

    Parameters
    ----------
    x0 : np.ndarray
        Initial point with shape (n,).
    A : np.ndarray
        Dictionary with shape (m,n).
    y : np.ndarray
        Target observation with shape (m,).
    l1 : float, positive
        L1 regularization parameter.
    l2 : float, positive
        L2 regularization parameter.
    params : ProjgradParams
        Optional parameters for the algorithm (see `utils.ProjgradParams`).

    Returns
    -------
    xopt : np.ndarray
        Solution of the problem with shape (n,).
    monitor : ProjgradMonitor
        Monitoring values (see `utils.ProjgradMonitor`).
    """

    validate_inputs(x0, A, y, l1, l2)
    params.validate()

    if params.verbosity:
        print_header(A, l1, l2, params)

    monitor = ProjgradMonitor()

    ###########################   Initialization   #############################

    # /!\ No FLOPs are counted for the initialization as all values can be 
    # /!\ computed off-line

    start_time = time.time()
    it = 0

    # Precomputed values
    m, n    = A.shape
    ninit   = n
    sqrt_l2 = np.sqrt(l2)
    Aty     = A.T @ y
    Aty_l1  = Aty - l1
    AtA     = A.T @ A
    gram    = AtA + l2 * np.eye(n)
    L       = np.max(np.abs(np.linalg.eigvals(gram)))

    # Iterates
    x       = np.maximum(x0,0)
    u       = y - A @ x
    xold    = np.copy(x)
    zold    = np.copy(x)
    told    = 1

    # Best gap recorded so far for screening and relaxing tests
    best_x   = np.copy(x)
    best_Atu = np.copy(A.T @ u)
    best_gap = duality_gap(x,u,A,y,l1,l2)

    # Screened and relaxed coefficients
    idx  = np.arange(n)                     # Initial indices
    S    = np.zeros(n,dtype=bool)           # Indices relaxed
    S[params.Sinit] = True                  # Indices relaxed at initialization
    Sbar = np.invert(S)                     # Indices not fixed yet
    sc   = set()                            # Screened coefficients
    rc   = set()                            # Relaxed coefficients
    rc.update(idx[S])
    
    # Data of the transformed problem
    Ginv = np.linalg.inv(gram[:,S][S,:])    # (A[:,S].T @ A[:,S] + l2 * I)^{-1}
    b    = Ginv @ Aty_l1[S]                 # Ginv @ Aty_l1[S]
    B    = - Ginv @ A[:,S].T @ A[:,Sbar]    # - Ginv @ AtA[S,Sbar]
    yw   = y - A[:,S] @ b                   # y - A[:,S] @ b
    Aw   = A[:,Sbar] + A[:,S] @ B           # A[:,Sbar] + A[:,S] @ B
    tw   = l1 + B.T @ (l1 + l2 * b)         # l1 + B.T @ (l1 + l2 * b)

    # Monitoring
    monitor.obj.append(primal_value(x,A,y,l1,l2))
    monitor.gap.append(duality_gap(x,u,A,y,l1,l2))
    monitor.flops.append(0)
    monitor.sc_size.append(len(sc))
    monitor.rc_size.append(len(rc))
    if params.coherence:
        monitor.coherence.append(coherence(Aw))
    else:
        monitor.coherence.append(np.nan)


    #############################   Main loop   ################################

    while True:

        it       += 1
        flops     = 0               # FLOPs of the current iteration
        S_size    = np.sum(S)       # ie. card(S)
        Sbar_size = np.sum(Sbar)    # ie. card(Sbar)

        # Iterates update
        Awz     = Aw @ zold[Sbar]
        u       = yw - Awz
        Awtu    = Aw.T @ u
        BBz     = B.T @ (B @ zold[Sbar])
        Awtu_tw = Awtu - tw
        grad    = Awtu_tw - l2 * (zold[Sbar] + BBz)
        x[Sbar] = np.maximum(zold[Sbar] + grad/L, 0)
        x[S]    = np.maximum(b + B @ x[Sbar],0)
        t       = (1 + np.sqrt(1 + 4 * told ** 2)) / 2
        z       = x + ((told - 1) / t) * (x - xold)

        # ------------------------- FLOPs count ------------------------------ #
        # Iterates update
        #   Awz         : 2m|Sbar|
        #   u           : m
        #   Awtu        : 2m|Sbar|
        #   BBz         : 4|Sbar||S|
        #   Awtu_tw     : |Sbar|
        #   grad        : 2|Sbar| + |S|
        #   x[Sbar]     : 2|Sbar|
        #   x[S]        : 2|S|^2 + |S|
        #   t           : 6
        #   z           : 3n + 2
        flops += 4*m*Sbar_size + 4*Sbar_size*S_size + 2*S_size**2 + 3*n + m + \
                 5*Sbar_size + 2*S_size + 8
        # -------------------------------------------------------------------- #
        
        # Gap computation
        zpos      = np.maximum(zold[Sbar],0)
        Atu       = np.zeros(n)
        Atu[S]    = A[:,S].T @ u
        Atu[Sbar] = Awtu - B.T @ Atu[S]
        y_u       = y - u
        v         = np.maximum(Atu-l1, 0)
        vpos      = v[v>0]
        pv        = 0.5*u@u + tw@zpos + (l2/2)*(zpos@BBz + zpos@zpos + b@b) + l1*np.sum(b)
        dv        = 0.5 * (1 - (y_u@y_u) - (1/l2)*(vpos@vpos))
        gap       = np.abs(pv-dv)

        # Update best primal and dual points recorded so far
        if gap < best_gap:
            best_gap = gap
            best_x  = np.copy(np.maximum(zold,0))
            best_Atu = np.copy(Atu)

        # ------------------------- FLOPs count ------------------------------ #
        # GAP computation
        #   Atu[S]      : 2m|S|
        #   Atu[Sbar]   : n + 2|S|m
        #   y_u         : m
        #   v           : n
        #   pv          : 2m + 6|Sbar| + 3|S| + 5
        #   dv          : 2m + 2|vpos| + 4
        #   gap         : 1
        flops += 4*m*S_size + 2*n + 5*m + 5*Sbar_size + 3*S_size + 10
        # -------------------------------------------------------------------- #


        if params.screen:

            # Screening test
            radius = np.sqrt(2 * best_gap)
            screen  = np.where(best_Atu <= l1 - radius)[0]
            sc.update(idx[screen])

            # Dimension reduction
            if len(screen) > 0:
                screen_Sbar = np.where(best_Atu[Sbar] <= l1 - radius)[0]
                A        = np.delete(A, screen, axis=1)
                gram     = np.delete(gram, screen, axis=0)
                gram     = np.delete(gram, screen, axis=1)
                AtA      = np.delete(AtA, screen, axis=0)
                AtA      = np.delete(AtA, screen, axis=1)
                Atu      = np.delete(Atu, screen)
                Aty      = np.delete(Aty, screen)
                Aty_l1   = np.delete(Aty_l1, screen)
                x        = np.delete(x, screen)
                z        = np.delete(z, screen)
                best_x   = np.delete(best_x, screen)
                best_Atu = np.delete(best_Atu, screen)
                idx      = np.delete(idx, screen)
                S        = np.delete(S, screen)
                Sbar     = np.delete(Sbar, screen)
                Aw       = np.delete(Aw, screen_Sbar, axis=1)
                B        = np.delete(B, screen_Sbar, axis=1)
                tw       = np.delete(tw, screen_Sbar)
                _, n     = A.shape

            # ----------------------- FLOPs count ---------------------------- #
            # Screening
            #   radius : 2
            #   screen : 1
            flops += 3
            # ---------------------------------------------------------------- #

        if params.relax:

            # Relaxing test
            radius    = np.sqrt(2 * best_gap) / sqrt_l2
            relax_Atu = np.where(best_Atu > l1 + radius * sqrt_l2)[0]
            relax_x   = np.where(best_x > radius)[0]
            relax     = np.unique(np.hstack((relax_Atu, relax_x)))
            rc.update(idx[relax])

            # ----------------------- FLOPs count ---------------------------- #
            # Relaxing
            #   radius      : 2
            #   relax_Atu   : 2
            flops += 4
            # ---------------------------------------------------------------- #

            # Update problem data
            if len(relax) > 0:
                for i in relax[Sbar[relax]]:
                    Ginv, B, Aw, b, yw, tw = update_data(Ginv, B, Aw, b, yw, tw,
                                                        i, A, y, S, Sbar, l1, 
                                                        l2, AtA, gram, Aty_l1
                                                        )

                    # -------------------- FLOPs count ----------------------- #
                    # Update problem data
                    #   See algebra.update_data for FLOPs count detail
                    flops += 1.5*S_size**2 + S_size*Sbar_size + m*S_size + \
                             9*S_size + Sbar_size + 14
                    # -------------------------------------------------------- #

                    S[i]       = True
                    Sbar[i]    = False
                    S_size    += 1
                    Sbar_size += 1
            
                    
        if len(sc)+len(rc) == ninit:

            # Closed form for the optimizer
            x[S]    = b
            x[Sbar] = 0

            # /!\ Following operations are only used for monitoring purpose
            # /!\ but are not needed in practice because convergence is
            # /!\ already certified. Thus, no FLOPs are counted.
            monitor.obj.append(primal_value(x,A,y,l1,l2))
            monitor.gap.append(duality_gap(x,y-A@x,A,y,l1,l2))
            monitor.flops.append(monitor.flops[-1] + flops)
            monitor.sc_size.append(len(sc))
            monitor.rc_size.append(len(rc))
            monitor.stopcrit = StopCrit.CONVERGED
            if params.coherence:
                monitor.coherence.append(coherence(Aw))
            else:
                monitor.coherence.append(np.nan)
            if params.verbosity and it % params.showevery == 0: 
                print_trace(it, monitor)
            break
            

        # Displays and monitoring
        monitor.obj.append(pv)
        monitor.gap.append(gap)
        monitor.flops.append(monitor.flops[-1] + flops)
        monitor.sc_size.append(len(sc))
        monitor.rc_size.append(len(rc))
        if params.coherence:
            monitor.coherence.append(coherence(Aw))
        else:
            monitor.coherence.append(np.nan)

        if params.verbosity and it % params.showevery == 0: 
            print_trace(it, monitor)

        if gap <= params.tol:
            monitor.stopcrit = StopCrit.CONVERGED
            break
        elif it+1 >= params.maxit:
            monitor.stopcrit = StopCrit.INTERRUPTED
            break

        xold = np.copy(x)
        zold = np.copy(z)
        told = np.copy(t)


    ############################## Postprocessing ##############################

    monitor.sc = sc
    monitor.rc = rc
    monitor.solvetime = time.time() - start_time

    xopt      = np.zeros(ninit)
    xopt[idx] = x

    if params.verbosity:
        print_footer(monitor)
        
    return xopt, monitor
